/******************************************************************************
** array.c
** =======

Array Utility Routines
James Conway, LMES/IBS, Grenoble, France (email: James.Conway@ibs.fr)

06-Mar-2000: Add CopyArrayByteToFloat, CopyArrayFloatToByte
             Rename old CopyArrayFloatToByte to CopyArrayFloatToUByte
             Add GetStatsByteArray, rename old GetStatsByteArray to GetStatsUByteArray
07-Mar-2000: Add GetStatsWordArray and GetStatsUWordArray
09-Mar-2000: Add GetStatsFloatArray3, change "char" to "signed char"

double SumDoubleArray  (double *aptr, int count);
float  SumFloatArray   (float  *aptr, int count);
void   ZeroDoubleArray (double *aptr, int count);
void   ZeroFloatArray  (float  *aptr, int count);
double AvgDoubleArray  (double *aptr, int count);
float  AvgFloatArray   (float  *aptr, int count);

void   CopyArrayByteToFloat  (   signed char *src, float *dst, long size );
void   CopyArrayUByteToFloat ( unsigned char *src, float *dst, long size );
void   CopyArrayFloatToUByte ( float *src, unsigned char *dst, long size );
void   CopyArrayFloatToByte  ( float *src,   signed char *dst, long size );
void   CopyArrayShortToFloat ( short *src, float *dst, long size );
void   CopyArrayFloatToShort ( float *src, short *dst, long size );
void   CopyArrayFloatToFloat ( float *src, float *dst, long size );

void   AddFloatArrays        ( float  *sum,  float  *a1, float  *a2, long count );
void   AddDoubleArrays       ( double *sum,  double *a1, double *a2, long count );
void   AddFloatToDoubleArray ( double *sum,  float  *a1, long count );
void   SubFloatArrays        ( float  *diff, float  *a1, float  *a2, long count );

void   ExpandAndPadArray     ( float *array, long dx0, long dy0, long dx1, long dy1; float padValue );
void   ShrinkFloatArray      ( float *array, long dx0, long dy0, long dx1, long dy1 );
void   SubtractMeanFromArray ( float *array, long size );
void   HistoFloatArray       ( float *array, long size, double *mean, double *bkgnd );
void   ScaleDoubleArray      ( double *array, long size, double scale, double offset );
void   ScaleFloatArray       ( float *array, long size, float scale, float offset );
void   MinMaxFloatArray      ( float *array, long size, float *min, float *max );
void   GetStatsByteArray     (   signed char *array, long size,   signed char *min,   signed char *max, float *mean, float *stdv );
void   GetStatsUByteArray    ( unsigned char *array, long size, unsigned char *min, unsigned char *max, float *mean, float *stdv );
void   GetStatsWordArray     ( short *array, long size, short *min, short *max, float *mean, float *stdv );
void   GetStatsUWordArray    ( unsigned short *array, long size, unsigned short *min, unsigned short *max, float *mean, float *stdv );
void   GetStatsFloatArray    ( float *array, long size, float *min, float *max, float *mean, float *stdv );
void   GetStatsFloatArray2   ( float *array, long dx, long dy, float radius, float *min, float *max, float *mean, float *stdv );
void   GetStatsFloatArray3   ( float *array, long dx, long dy, float radius, float *min, float *max, float *mean, float *stdv );
void   SetFloatArrayMean     ( float *array, long size, float newMean );

void   Regress               ( float *data, float *signal, long size, double *a, double *b );
void   CalcMeanVariance      ( float *array, long size, double *average, double *variance, float maskVal );

******************************************************************************/

#include <math.h>
#include "array.h"

/******************************************************************************
** SumDoubleArray
** ==============
*/
double SumDoubleArray(double *aptr, int count)
{
  double sum = 0.0;

  while(count-- > 0)
    sum += *aptr++;
  return(sum);
}
/******************************************************************************
** SumFloatArray
** =============
Be careful of underflow!
*/
float SumFloatArray(float *aptr, int count)
{
  double sum = 0.0;

  while(count-- > 0)
    sum += (double) *aptr++;
  return((float) sum);
}
/******************************************************************************
** ZeroDoubleArray
** ===============
*/
void ZeroDoubleArray(double *aptr, int count)
{
  while(count--)
    *aptr++ = 0.0;
}
/******************************************************************************
** ZeroFloatArray
** ==============
*/
void ZeroFloatArray(float *aptr, int count)
{
  while(count--)
    *aptr++ = 0.0;
}
/******************************************************************************
** AvgDoubleArray
** ==============
*/
double AvgDoubleArray(double *aptr, int count)
{
  return(SumDoubleArray(aptr, count) / (double) count);
}
/******************************************************************************
** AvgFloatArray
** =============
*/
float AvgFloatArray(float *aptr, int count)
{
  return(SumFloatArray(aptr, count) / (double) count);
}
/******************************************************************************
** CopyArrayByteToFloat
** ====================
Copy array of signed byte to array of float. Arrays can be overlapping as
long as smaller (byte) data is at end of array space. 
*/
void CopyArrayByteToFloat(signed char *src, float *dst, long size)
{
  signed char *s;
  float *d;

  s = src + size;
  d = dst + size;
  while(s>src)
    *(--d) = (float) *(--s);
}
/******************************************************************************
** CopyArrayUByteToFloat
** =====================
Copy array of unsigned byte to array of float. Arrays can be overlapping as
long as smaller (byte) data is at end of array space. 
*/
void CopyArrayUByteToFloat(unsigned char *src, float *dst, long size)
{
  unsigned char *s;
  float *d;

  s = src + size;
  d = dst + size;
  while(s>src)
    *(--d) = (float) *(--s);
}
/******************************************************************************
** CopyArrayFloatToByte
** ====================
Copy array of float to array of signed byte. Arrays can be overlapping as
long as smaller (short) data is at beginning of array space.
*/
void CopyArrayFloatToByte(float *src, signed char *dst, long size)
{
  float *s, temp;
  signed char *d;

  s = src;
  d = dst;
  while((size--) > 0)  {
    temp = *(s++);
    if(temp<-128.0)
      *d = -128;
    else if(temp>127.0)
      *d = 127;
    else
      *d = (signed char) temp;
    d++;
  }
}
/******************************************************************************
** CopyArrayFloatToUByte
** =====================
Copy array of float to array of unsigned byte. Arrays can be overlapping as
long as smaller (short) data is at beginning of array space.
*/
void CopyArrayFloatToUByte(float *src, unsigned char *dst, long size)
{
  float *s, temp;
  unsigned char *d;

  s = src;
  d = dst;
  while((size--) > 0)  {
    temp = *(s++);
    if(temp<0.0)
      *d = 0;
    else if(temp>255.0)
      *d = 255;
    else
      *d = (unsigned char) temp;
    d++;
  }
}
/******************************************************************************
** CopyArrayShortToFloat
** =====================
Copy array of short to array of float. Arrays can be overlapping as long as
smaller (short) data is at beginning of array space. 
*/
void CopyArrayShortToFloat(short *src, float *dst, long size)
{
  short *s;
  float *d;

  s = src + size;
  d = dst + size;
  while(s>src)
    *(--d) = (float) *(--s);
}
/******************************************************************************
** CopyArrayFloatToShort
** =====================
Copy array of float to array of short. Arrays can be overlapping as long as
smaller (short) data is at beginning of array space. 
*/
void CopyArrayFloatToShort(float *src, short *dst, long size)
{
  float *s;
  short *d;

  s = src;
  d = dst;
  while((size--) > 0)
    *(d++) = (short) *(s++);
}
/******************************************************************************
** CopyArrayFloatToFloat
** =====================
Copy array of float to array of float. 
*/
void CopyArrayFloatToFloat(float *src, float *dst, long size)
{
  float *s;
  float *d;

  s = src;
  d = dst;
  while((size--) > 0)
    *d++ = *s++;
}
/******************************************************************************
** AddFloatArrays
** ==============
*/
void AddFloatArrays(float *sum, float *a1, float *a2, long count)
{
  while(--count >= 0)
    *sum++ = (*a1++) + (*a2++);
}
/******************************************************************************
** AddDoubleArrays
** ===============
*/
void AddDoubleArrays(double *sum, double *a1, double *a2, long count)
{
  while(--count >= 0)
    *sum++ = (*a1++) + (*a2++);
}
/******************************************************************************
** AddFloatToDoubleArray
** =====================
*/
void AddFloatToDoubleArray(double *sum, float *a1, long count)
{
  while(--count >= 0)
    *sum++ += (double) (*a1++);
}
/******************************************************************************
** AddFloatToFloatArray
** ====================
Add array of float to array of float. 
*/
void AddFloatToFloatArray(float *src, float *dst, long size)
{
  float *s;
  float *d;

  s = src;
  d = dst;
  while((size--) > 0)
    *d++ += *s++;
}
/******************************************************************************
** SubFloatArrays
** ==============
*/
void SubFloatArrays(float *diff, float *a1, float *a2, long count)
{
  while(--count >= 0)
    *diff++ = (*a1++) - (*a2++);
}
/******************************************************************************
** ExpandAndPadArray
** =================
Small array dx0*dy0 is packed into large array dx1*dy1 from begining. Unpack
and pad end of rows and extra rows with padValue
*/
void ExpandAndPadArray( float *array
                      , long dx0, long dy0
                      , long dx1, long dy1
                      , float padValue
                      )
{
  float *src, *dst;
  int    x, y;

  if(dx0==dx1 && dy0==dy1)      // same size - skip
    return;

  dst = array + (dx1 * dy1);
  for(y=dy1-1; y>=dy0; y--)
    for(x=dx1-1; x>=0; x--)
      *(--dst) = padValue;

  src = array + (dx0 * dy0);
  for(y=dy0-1; y>=0; y--)  {
    for(x=dx1-1; x>=dx0; x--)
      *(--dst) = padValue;
    for(x=dx0-1; x>=0; x--)
      *(--dst) = *(--src);
  }
}
/******************************************************************************
** ShrinkFloatArray
** ================
Small array dx0*dy0 is trimmed from large array dx1*dy1 and put at begining.
*/
void ShrinkFloatArray(float *array, long dx0, long dy0, long dx1, long dy1)
{
  float *src, *dst;
  int    x, y;

  dst = array;
  for(y=0; y<dy0; y++)  {
    src = array + dx1*y;        // beginning of new line in larger array
    for(x=0; x<dx0; x++)
      *(dst++) = *(src++);
  }
}
/******************************************************************************
** SubtractMeanFromArray
** =====================
*/
void SubtractMeanFromArray(float *array, long size)
{
  long    s;
  float  *a;
  double  sum;
  float   mean;

  sum = 0.0;
  a = array;
  for(s=0; s<size; s++)
    sum += (double) *a++;
  mean = (float) (sum / (double) size);

  a = array;
  for(s=0; s<size; s++)
    *a++ -= mean;
}
/******************************************************************************
** HistoFloatArray
** ===============
*/
void HistoFloatArray(float *array, long size, double *mean, double *bkgnd)
{
  double sum=0.0;
  long   count=0;
  float  fBkgnd;

  fBkgnd = *array;            // (0,0) element is part of constant background
  *bkgnd = (double) *array;

  while(--size >= 0)  {
    if(*array != fBkgnd)  {
      sum   += (double) *array;
      count += 1;
    }
    array++;
  }
  if(count == 0)
    *mean = 0.0;
  else
    *mean = sum / (double) count;
}
/******************************************************************************
** ScaleDoubleArray
** ================
*/
void ScaleDoubleArray(double *array, long size, double scale, double offset)
{
  while(--size >= 0)  {
    *array *= scale;
    *array += offset;
    array++;
  }
}
/******************************************************************************
** ScaleFloatArray
** ===============
*/
void ScaleFloatArray(float *array, long size, float scale, float offset)
{
  while(--size >= 0)  {
    *array *= scale;
    *array += offset;
    array++;
  }
}
/******************************************************************************
** MinMaxFloatArray
** ================
*/
void MinMaxFloatArray(float *array, long size, float *min, float *max)
{
  *min = *max = *array;
  while(--size >= 0)  {
    if(*array < *min)  *min = *array;
    else if(*array > *max)  *max = *array;
    array++;
  }
}
/******************************************************************************
** GetStatsByteArray
** =================
*/
void GetStatsByteArray( signed char *array, long size
                      , signed char  *min,  signed char  *max
                      , float *mean, float *stdv
                      )
{
  long i;
  signed long suml, sumh;
  unsigned long sum2l, sum2h;
  double sum, sum2;
  signed char lmin, lmax;

  suml  = 0;    // sum - lo 4 bytes
  sumh  = 0;    // sum - hi 4 bytes
  sum2l = 0;    // sum squared - lo 4 bytes
  sum2h = 0;    // sum squared - hi 4 bytes

  *min  = *max  = 0;
  *mean = *stdv = 0.0;    // defaults in case of size=0 or 1

  if(size<1)
    return;

  lmin = lmax = *array;

  for(i=0; i<size; i++)  {
    short value = *array++;
    if     (value < lmin)  lmin = value;
    else if(value > lmax)  lmax = value;
    suml  = suml  + value;
    sum2l = sum2l + value*value;
    if(suml>2000000000)  {        // avoid overflow of 4-byte integer
      sumh = sumh + 1;
      suml = suml - 2000000000;
    }
    else if(suml<-2000000000)  {  // avoid overflow of 4-byte integer
      sumh = sumh - 1;
      suml = suml + 2000000000;
    }
    if(sum2l>2000000000)  {       // avoid overflow of 4-byte integer
      sum2h = sum2h + 1;
      sum2l = sum2l - 2000000000;
    }
  }

  sum  = 2e9*((double) sumh)  + (double) suml;
  sum2 = 2e9*((double) sum2h) + (double) sum2l;

  *min = lmin;
  *max = lmax;

  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsUByteArray
** ==================
*/
void GetStatsUByteArray( unsigned char *array, long size
                       , unsigned char *min, unsigned char *max
                       , float *mean, float *stdv
                       )
{
  long i;
  unsigned long suml, sumh, sum2l, sum2h;
  double sum, sum2;
  unsigned char lmin, lmax;

  suml  = 0;    // sum - lo 4 bytes
  sumh  = 0;    // sum - hi 4 bytes
  sum2l = 0;    // sum squared - lo 4 bytes
  sum2h = 0;    // sum squared - hi 4 bytes

  *min  = *max  = 0;
  *mean = *stdv = 0.0;    // defaults in case of size=0 or 1

  if(size<1)
    return;

  lmin = lmax = *array;

  for(i=0; i<size; i++)  {
    unsigned short value = *array++;
    if     (value < lmin)  lmin = value;
    else if(value > lmax)  lmax = value;
    suml  = suml  + value;
    sum2l = sum2l + value*value;
    if(suml>2000000000)  {        // avoid overflow of 4-byte integer
      sumh = sumh + 1;
      suml = suml - 2000000000;
    }
    if(sum2l>2000000000)  {       // avoid overflow of 4-byte integer
      sum2h = sum2h + 1;
      sum2l = sum2l - 2000000000;
    }
  }

  sum  = 2e9*((double) sumh)  + (double) suml;
  sum2 = 2e9*((double) sum2h) + (double) sum2l;

  *min = lmin;
  *max = lmax;

  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsWordArray
** =================
*/
void GetStatsWordArray( short *array, long size
                      , short *min, short *max
                      , float *mean, float *stdv
                      )
{
  long i;
  signed long suml, sumh;
  unsigned long sum2l, sum2h;
  double sum, sum2;
  short lmin, lmax;

  suml  = 0;    // sum - lo 4 bytes
  sumh  = 0;    // sum - hi 4 bytes
  sum2l = 0;    // sum squared - lo 4 bytes
  sum2h = 0;    // sum squared - hi 4 bytes

  *min  = *max  = 0;
  *mean = *stdv = 0.0;    // defaults in case of size=0 or 1

  if(size<1)
    return;

  lmin = lmax = *array;

  for(i=0; i<size; i++)  {
    long value = *array++;
    if     (value < lmin)  lmin = value;
    else if(value > lmax)  lmax = value;
    suml  = suml  + value;
    sum2l = sum2l + value*value;
    if(suml>2000000000)  {  // avoid overflow of 4-byte integer
      sumh = sumh + 1;
      suml = suml - 2000000000;
    }
    else if(suml<-2000000000)  {  // avoid overflow of 4-byte integer
      sumh = sumh - 1;
      suml = suml + 2000000000;
    }
    if(sum2l>2000000000)  {  // avoid overflow of 4-byte integer
      sum2h = sum2h + 1;
      sum2l = sum2l - 2000000000;
    }
  }

  sum  = 2e9*((double) sumh)  + (double) suml;
  sum2 = 2e9*((double) sum2h) + (double) sum2l;

  *min = lmin;
  *max = lmax;

  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsUWordArray
** ==================
*/
void GetStatsUWordArray( unsigned short *array, long size
                       , unsigned short *min, unsigned short *max
                       , float *mean, float *stdv
                       )
{
  long i;
  unsigned long suml, sumh, sum2l, sum2h;
  double sum, sum2;
  unsigned short lmin, lmax;

  suml  = 0;    // sum - lo 4 bytes
  sumh  = 0;    // sum - hi 4 bytes
  sum2l = 0;    // sum squared - lo 4 bytes
  sum2h = 0;    // sum squared - hi 4 bytes

  *min  = *max  = 0;
  *mean = *stdv = 0.0;    // defaults in case of size=0 or 1

  if(size<1)
    return;

  lmin = lmax = *array;

  for(i=0; i<size; i++)  {
    unsigned long value = *array++;
    if     (value < lmin)  lmin = value;
    else if(value > lmax)  lmax = value;
    suml  = suml  + value;
    sum2l = sum2l + value*value;
    if(suml>2000000000)  {  // avoid overflow of 4-byte integer
      sumh = sumh + 1;
      suml = suml - 2000000000;
    }
    if(sum2l>2000000000)  {  // avoid overflow of 4-byte integer
      sum2h = sum2h + 1;
      sum2l = sum2l - 2000000000;
    }
  }

  sum  = 2e9*((double) sumh)  + (double) suml;
  sum2 = 2e9*((double) sum2h) + (double) sum2l;

  *min = lmin;
  *max = lmax;

  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsFloatArray
** ==================
*/
void GetStatsFloatArray( float *array, long size
                       , float *min, float *max
                       , float *mean, float *stdv
                       )
{
  double sum=0.0, sum2=0.0;
  long lsize = size;

  *min = *max = *mean = *stdv = 0.0;    // defaults in case of error
  if(size<1)
    return;
  *min = *max = *array;
  while(--lsize >= 0)  {
    if(*array < *min)  *min = *array;
    else if(*array > *max)  *max = *array;
    sum  += *array;
    sum2 += (*array) * (*array);
    array++;
  }
  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsFloatArray2
** ===================
Same as above, but uses values within a radius centered in the array.
*/
void GetStatsFloatArray2( float *array, long dx, long dy, float radius
                        , float *min, float *max, float *mean, float *stdv)
{
  double sum=0.0, sum2=0.0;
  int x, y;
  float x2, y2;
  const float xcen = ((float)(dx-1))/2.0, ycen = ((float)(dy-1))/2.0;
  const float rad2 = radius * radius;
  const long size = dx * dy;

  *min = *max = *mean = *stdv = 0.0;    // defaults in case of error
  if(size<1)
    return;

  *min = *max = *array;
  for(y=0; y<dy; y++)  {
    y2 = ((float) y) - ycen;
    y2 *= y2;
    for(x=0; x<dx; x++, array++)  {
      x2 = ((float) x) - xcen;
      x2 *= x2;
      if(x2+y2 < rad2)  {
        if(*array < *min)  *min = *array;
        else if(*array > *max)  *max = *array;
        sum  += *array;
        sum2 += (*array) * (*array);
      }
    }
  }
  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** GetStatsFloatArray3
** ===================
Same as above, but uses values *outside* a radius centered in the array.
*/
void GetStatsFloatArray3( float *array, long dx, long dy, float radius
                        , float *min, float *max, float *mean, float *stdv)
{
  double sum=0.0, sum2=0.0;
  int x, y;
  float x2, y2;
  const float xcen = ((float)(dx-1))/2.0, ycen = ((float)(dy-1))/2.0;
  const float rad2 = radius * radius;
  const long size = dx * dy;

  *min = *max = *mean = *stdv = 0.0;    // defaults in case of error
  if(size<1)
    return;

  *min = *max = *array;
  for(y=0; y<dy; y++)  {
    y2 = ((float) y) - ycen;
    y2 *= y2;
    for(x=0; x<dx; x++, array++)  {
      x2 = ((float) x) - xcen;
      x2 *= x2;
      if(x2+y2 > rad2)  {
        if(*array < *min)  *min = *array;
        else if(*array > *max)  *max = *array;
        sum  += *array;
        sum2 += (*array) * (*array);
      }
    }
  }
  *mean = sum / (double) size;
  if(size>1)
    *stdv = sqrt(  (sum2 - sum*sum/(double) size) / (double) (size-1)  );
}
/******************************************************************************
** SetFloatArrayMean
** =================
*/
void SetFloatArrayMean(float *array, long size, float newMean)
{
  float oldMean = AvgFloatArray(array, size);
  ScaleFloatArray(array, size, 1.0, newMean-oldMean);
}
/******************************************************************************
** Regress
** =======
Task:    compute a linear regression. Calculate a and b so that
             |a*data + b - signal| is minimum
Comment: the 2 images, data and signal, have the same size
Input:   data and signal
         nx, ny and nz, the images' dimensions
Output:  adjusted = a*data - b
         a and b
*/
void Regress ( float *data, float *signal, long size
             , double *a, double *b
             )
{
  float *dptr, *sptr;
  long k;
  double Dmean, Smean, numA, denA;
  double x, y;

  dptr = data;
  sptr = signal;
  Dmean = Smean = 0.0;
  for(k=0; k<size; k++)  {
    Dmean += (double) *dptr++;
    Smean += (double) *sptr++;
  }
  Dmean /= (double) size;
  Smean /= (double) size;

  dptr = data;
  sptr = signal;
  numA = denA = 0.0;
  for(k=0; k<size; k++)  {
    x = ((double) *dptr++) - Dmean;
    y = ((double) *sptr++) - Smean;
    numA += (x*y);
    denA += (x*x);
  }

  x = numA/denA;                // b = Smean - a*Nmean;
  y = (Dmean - (Smean/x));
  *a = x;
  *b = -y*x;

//   for(k=0; k<(nx*ny*nz); k++)
//     *aptr++ = (float) (x * ((double) *dptr++) - y*x);
}
/******************************************************************************
** CalcMeanVariance
** ================
*/
void CalcMeanVariance(float *array, long size
                , double *average, double *variance, float maskVal)
{
  double sum=0, sum2=0, value;
  long p;
  unsigned long count=0;

  for(p=0; p<size; p++, array++)
    if(*array != maskVal)  {
      value = (double) *array;
      sum  += value;
      sum2 += value * value;
      count++;
    }
  if(count>0)
    *average = sum / (double) (count);
  else
    *average = 0.0;

  if(count>1)
    *variance = ( sum2 - sum*sum/(double) (count) ) / (double) (count-1);
  else
    *variance = 0;
}
/*****************************************************************************/
